#
# Duties for a Formulation
#
abstract type AbstractFormDuty end
abstract type AbstractMasterDuty <: AbstractFormDuty end
abstract type AbstractSpDuty <: AbstractFormDuty end

"Formulation provided by the user."
struct Original <: AbstractFormDuty end

"Master of a formulation decomposed using Dantzig-Wolfe."
struct DwMaster <: AbstractMasterDuty end

"Master of a formulation decomposed using Benders."
struct BendersMaster <: AbstractMasterDuty end

"A pricing subproblem of a formulation decomposed using Dantzig-Wolfe."
mutable struct DwSp <: AbstractSpDuty 
    setup_var::Union{VarId, Nothing}
    lower_multiplicity::Int
    upper_multiplicity::Int
    column_var_kind::VarKind
end

"A Benders subproblem of a formulation decomposed using Benders."
struct BendersSp <: AbstractSpDuty 
    slack_to_first_stage::Dict{VarId, VarId}
end

BendersSp() = BendersSp(Dict{VarId, VarId}())

#
# Duties tree for a Variable
#
@exported_nestedenum begin
    Duty{Variable}
        AbstractOriginalVar <= Duty{Variable}
            OriginalVar <= AbstractOriginalVar
            #OriginalExpression <= AbstractOriginalVar
        AbstractMasterVar <= Duty{Variable}
            AbstractOriginMasterVar <= AbstractMasterVar
                MasterPureVar <= AbstractOriginMasterVar
                MasterBendFirstStageVar <= AbstractOriginMasterVar
            AbstractAddedMasterVar <= AbstractMasterVar
                MasterCol <= AbstractAddedMasterVar
                MasterArtVar <= AbstractAddedMasterVar
                MasterBendSecondStageCostVar <= AbstractAddedMasterVar
            AbstractImplicitMasterVar <= AbstractMasterVar
                AbstractMasterRepDwSpVar <= AbstractImplicitMasterVar
                    MasterRepPricingVar <= AbstractMasterRepDwSpVar
                    MasterRepPricingSetupVar <= AbstractMasterRepDwSpVar
        AbstractDwSpVar <= Duty{Variable}
            DwSpPricingVar <= AbstractDwSpVar
            DwSpSetupVar <= AbstractDwSpVar
            #DwSpPureVar <= AbstractDwSpVar
            DwSpPrimalSol <= AbstractDwSpVar
        AbstractBendSpVar <= Duty{Variable}
            AbstractBendSpSlackMastVar <= AbstractBendSpVar
                BendSpSlackFirstStageVar <= AbstractBendSpSlackMastVar
                    BendSpPosSlackFirstStageVar <= BendSpSlackFirstStageVar
                    BendSpNegSlackFirstStageVar <= BendSpSlackFirstStageVar
                BendSpSlackSecondStageCostVar <= AbstractBendSpSlackMastVar
            BendSpSepVar <= AbstractBendSpVar
            #BendSpPureVar <= AbstractBendSpVar
            BendSpPrimalSol <= AbstractBendSpVar
end

#
# Duties tree for a Constraint
#
@exported_nestedenum begin
    Duty{Constraint}
        AbstractOriginalConstr <= Duty{Constraint}
            OriginalConstr <= AbstractOriginalConstr
        AbstractMasterConstr <= Duty{Constraint}
            AbstractMasterOriginConstr <= AbstractMasterConstr
                MasterPureConstr <= AbstractMasterOriginConstr
                MasterMixedConstr <= AbstractMasterOriginConstr
            AbstractMasterAddedConstr <= AbstractMasterConstr
                MasterConvexityConstr <= AbstractMasterAddedConstr
                #MasterSecondStageCostConstr <= AbstractMasterAddedConstr
            #AbstractMasterImplicitConstr <= AbstractMasterConstr
                #AbstractMasterRepBendSpConstr <= AbstractMasterImplicitConstr
                    #MasterRepBendSpSecondStageCostConstr <= AbstractMasterRepBendSpConstr
                    #MasterRepBendSpTechnologicalConstr <= AbstractMasterRepBendSpConstr
            AbstractMasterCutConstr <= AbstractMasterConstr
                MasterBendCutConstr <= AbstractMasterCutConstr
                MasterUserCutConstr <= AbstractMasterCutConstr
            AbstractMasterBranchingConstr <= AbstractMasterConstr
                MasterBranchOnOrigVarConstr <= AbstractMasterBranchingConstr
        AbstractDwSpConstr <= Duty{Constraint}
            DwSpPureConstr <= AbstractDwSpConstr
            # <= AbstractDwSpConstr
            #DwSpRepMastBranchConstr <= AbstractDwSpConstr
        #AbstractBendSpPureConstr <= Duty{Constraint}
        AbstractBendSpConstr <= Duty{Constraint}
            AbstractBendSpMasterConstr <= AbstractBendSpConstr
                BendSpSecondStageCostConstr <= AbstractBendSpMasterConstr
                BendSpTechnologicalConstr <= AbstractBendSpMasterConstr
            BendSpPureConstr <= AbstractBendSpConstr
            BendSpDualSol <= AbstractBendSpConstr
end

#
# Methods to get extra information about duties
#
function isaStaticDuty(duty::NestedEnum)
    return duty <= OriginalVar ||
    #duty <= OriginalExpression ||
    duty <= MasterPureVar ||
    duty <= MasterArtVar ||
    duty <= MasterBendSecondStageCostVar ||
    duty <= MasterBendFirstStageVar ||
    duty <= MasterRepPricingVar ||
    duty <= MasterRepPricingSetupVar ||
    duty <= DwSpPricingVar ||
    duty <= DwSpSetupVar ||
    #duty <= DwSpPureVar ||
    duty <= DwSpPrimalSol ||
    duty <= DwSpDualSol ||
    duty <= BendSpSepVar ||
    #duty <= BendSpPureVar ||
    duty <= BendSpSlackFirstStageVar  ||
    duty <= BendSpSlackSecondStageCostVar ||
    duty <= OriginalConstr ||
    duty <= MasterPureConstr ||
    duty <= MasterMixedConstr ||
    duty <= MasterConvexityConstr ||
    #duty <= MasterSecondStageCostConstr ||
    duty <= DwSpPureConstr ||
    duty <= BendSpPureConstr ||
    duty <= BendSpDualSol ||
    duty <= BendSpPrimalSol ||
    duty <= BendSpSecondStageCostConstr ||
    duty <= BendSpTechnologicalConstr
end

function isaDynamicDuty(duty::NestedEnum)
    duty <= MasterCol ||
    duty <= MasterBranchOnOrigVarConstr ||
    duty <= MasterBendCutConstr ||
    duty <= MasterBranchOnOrigVarConstr
    #duty <= DwSpRepMastBranchConstr ||
    #duty <= DwSpRepMastBranchConstr
end

function isanOriginalRepresentatives(duty::NestedEnum)
    duty <= MasterPureVar ||
    duty <= MasterRepPricingVar
end

function isanArtificialDuty(duty::NestedEnum)
    return duty <= MasterArtVar
end